#!/usr/bin/env python3.6

"""Module to test optics corr utils class."""

import unittest
from siriuspy import util
from siriuspy.opticscorr import utils
from siriuspy.opticscorr.utils import HandleConfigNameFile


PUB_INTERFACE = (
    'HandleConfigNameFile',
    'get_default_config_name'
)


class TestOpticsCorrUtils(unittest.TestCase):
    """Test opticscorr utils class."""

    def test_public_interface(self):
        """Test module's public interface."""
        valid = util.check_public_interface_namespace(
            utils, PUB_INTERFACE, print_flag=True)
        self.assertTrue(valid)

    def test_HandleConfigNameFile(self):
        """Test HandleConfigNameFile class."""
        with self.assertRaises(ValueError):
            _ = HandleConfigNameFile('', 'tune')
        with self.assertRaises(ValueError):
            _ = HandleConfigNameFile('SI', '')

    def test_set_config_name(self):
        """Test set_config_name function."""
        cn_handler = HandleConfigNameFile('SI', 'tune')
        with self.assertRaises(TypeError):
            cn_handler.set_config_name(1)


if __name__ == "__main__":
    unittest.main()



"""Waveform classes."""

import numpy as _np

from mathphys import constants as _c
from mathphys import units as _mu

from ..pwrsupply.csdev import DEF_WFMSIZE_OTHERS as \
    _DEF_WFMSIZE_OTHERS
from ..search import MASearch as _MASearch

from . import util as _ru
from .magnet import get_magnet as _get_magnet


class WaveformParam:
    """Dipole parameterized Waveforms."""

    N = _np.array([0, 1, 2, 3])

    def __init__(
            self,
            duration=_ru.DEFAULT_PS_RAMP_DURATION,
            rampup1_start_time=_ru.DEFAULT_PS_RAMP_RAMPUP1_START_TIME,
            rampup2_start_time=_ru.DEFAULT_PS_RAMP_RAMPUP2_START_TIME,
            rampdown_start_time=_ru.DEFAULT_PS_RAMP_RAMPDOWN_START_TIME,
            rampdown_stop_time=_ru.DEFAULT_PS_RAMP_RAMPDOWN_STOP_TIME,
            rampup_smooth_intvl=_ru.DEFAULT_PS_RAMP_RAMPUP_SMOOTH_INTVL,
            rampdown_smooth_intvl=_ru.DEFAULT_PS_RAMP_RAMPDOWN_SMOOTH_INTVL,
            start_value=_ru.DEFAULT_PS_RAMP_START_CURRENT,
            rampup1_start_value=_ru.DEFAULT_PS_RAMP_RAMPUP1_START_CURRENT,
            rampup2_start_value=_ru.DEFAULT_PS_RAMP_RAMPUP2_START_CURRENT,
            rampdown_start_value=_ru.DEFAULT_PS_RAMP_RAMPDOWN_START_CURRENT,
            rampdown_stop_value=_ru.DEFAULT_PS_RAMP_RAMPDOWN_STOP_CURRENT,
            rampup_smooth_value=_ru.DEFAULT_PS_RAMP_RAMPUP_SMOOTH_CURRENT,
            rampdown_smooth_value=_ru.DEFAULT_PS_RAMP_RAMPDOWN_SMOOTH_CURRENT
                ):
        """Init method."""
        self._duration = duration
        self._start_value = start_value
        self._rampup1_start_time = rampup1_start_time
        self._rampup1_start_value = rampup1_start_value
        self._rampup_smooth_intvl = rampup_smooth_intvl
        self._rampup_smooth_value = rampup_smooth_value
        self._rampup2_start_time = rampup2_start_time
        self._rampup2_start_value = rampup2_start_value
        self._rampdown_smooth_intvl = rampdown_smooth_intvl
        self._rampdown_smooth_value = rampdown_smooth_value
        self._rampdown_start_time = rampdown_start_time
        self._rampdown_start_value = rampdown_start_value
        self._rampdown_stop_time = rampdown_stop_time
        self._rampdown_stop_value = rampdown_stop_value
        self._update()

    @property
    def duration(self):
        """Return duration."""
        return self._duration

    @property
    def start_value(self):
        """Return start_value."""
        return self._start_value

    @property
    def rampup1_start_time(self):
        """Return rampup1_start_time."""
        return self._rampup1_start_time

    @property
    def rampup1_start_value(self):
        """Return rampup1_start_value."""
        return self._rampup1_start_value

    @property
    def rampup2_start_time(self):
        """Return rampup2_start_time."""
        return self._rampup2_start_time

    @property
    def rampup2_start_value(self):
        """Return rampup2_start_value."""
        return self._rampup2_start_value

    @property
    def rampdown_start_time(self):
        """Return rampdown_start_time."""
        return self._rampdown_start_time

    @property
    def rampdown_start_value(self):
        """Return rampdown_start_value."""
        return self._rampdown_start_value

    @property
    def rampdown_stop_time(self):
        """Return rampdown_stop_time."""
        return self._rampdown_stop_time

    @property
    def rampdown_stop_value(self):
        """Return rampdown_stop_value."""
        return self._rampdown_stop_value

    @property
    def rampup_smooth_intvl(self):
        """Return rampup_smooth_intvl."""
        return self._rampup_smooth_intvl

    @property
    def rampup_smooth_value(self):
        """Return rampup_smooth_value."""
        return self._rampup_smooth_value

    @property
    def rampdown_smooth_intvl(self):
        """Return rampdown_smooth_intvl."""
        return self._rampdown_smooth_intvl

    @property
    def rampdown_smooth_value(self):
        """Return rampdown_smooth_value."""
        return self._rampdown_smooth_value

    @duration.setter
    def duration(self, value):
        """Set duration."""
        self._duration = value
        self._update()

    @start_value.setter
    def start_value(self, value):
        """Set start_value."""
        self._start_value = float(value)
        self._update()

    @rampup1_start_time.setter
    def rampup1_start_time(self, value):
        """Set rampup1_start_time."""
        self._rampup1_start_time = value
        self._update()

    @rampup1_start_value.setter
    def rampup1_start_value(self, value):
        """Set rampup1_start_value."""
        self._rampup1_start_value = float(value)
        self._update()

    @rampup2_start_time.setter
    def rampup2_start_time(self, value):
        """Set rampup2_start_time."""
        self._rampup2_start_time = value
        self._update()

    @rampup2_start_value.setter
    def rampup2_start_value(self, value):
        """Set rampup2_start_value."""
        self._rampup2_start_value = float(value)
        self._update()

    @rampdown_start_time.setter
    def rampdown_start_time(self, value):
        """Set rampdown_start_time."""
        self._rampdown_start_time = value
        self._update()

    @rampdown_start_value.setter
    def rampdown_start_value(self, value):
        """Set rampdown_start_value."""
        self._rampdown_start_value = float(value)
        self._update()

    @rampdown_stop_time.setter
    def rampdown_stop_time(self, value):
        """Set rampdown_stop_time."""
        self._rampdown_stop_time = value
        self._update()

    @rampdown_stop_value.setter
    def rampdown_stop_value(self, value):
        """Set rampdown_stop_value."""
        self._rampdown_stop_value = float(value)
        self._update()

    @rampup_smooth_intvl.setter
    def rampup_smooth_intvl(self, value):
        """Set rampup_smooth_intvl."""
        self._rampup_smooth_intvl = value
        self._update()

    @rampup_smooth_value.setter
    def rampup_smooth_value(self, value):
        """Set rampup_smooth_value."""
        self._rampup_smooth_value = value
        self._update()

    @rampdown_smooth_intvl.setter
    def rampdown_smooth_intvl(self, value):
        """Set rampdown_smooth_intvl."""
        self._rampdown_smooth_intvl = value
        self._update()

    @rampdown_smooth_value.setter
    def rampdown_smooth_value(self, value):
        """Set rampdown_smooth_value."""
        self._rampdown_smooth_value = value
        self._update()

    @property
    def rampup1_slope(self):
        """Return rampup1 slope."""
        return self._c[1][1]

    @property
    def rampup2_slope(self):
        """Return rampup2 slope."""
        return self._c[2][1]

    @property
    def rampdown_slope(self):
        """Return rampdown slope."""
        return self._c[3][1]

    def eval_at(self, time):
        """."""
        time = _np.array(time)
        value = _np.zeros(time.shape)
        stop0 = self._rampup1_start_time
        stop12 = 0.5 * (self._rampup2_start_time + self._rampdown_start_time)
        stop23 = self._rampdown_stop_time
        # region 0
        sel = (time <= stop0)
        if _np.any(sel):
            value[sel] = self._func_polynom(0, time[sel])
        # region 12
        sel = (time > stop0) & (time <= stop12)
        if _np.any(sel):
            value[sel] = self._func_region_12(time[sel])
        # region 23
        sel = (time > stop12) & (time <= stop23)
        if _np.any(sel):
            value[sel] = self._func_region_23(time[sel])
        # region 4
        sel = (time > stop23)
        if _np.any(sel):
            value[sel] = self._func_polynom(4, time[sel])
        return value

    def _update(self):
        self._t = []  # start and stop times
        self._v = []  # start amd stop values
        self._c = []  # polynomial coefficients

        # region 0 - cubic
        self._t.append(_np.array([0.0, self._rampup1_start_time]))
        self._v.append(_np.array([self._start_value,
                                  self._rampup1_start_value]))
        self._c.append([0.0, 0.0, 0.0, 0.0])
        # region 1 - linear rampup1
        time = _np.array([self._rampup1_start_time, self._rampup2_start_time])
        value = _np.array([self._rampup1_start_value,
                           self._rampup2_start_value])
        self._t.append(time)
        self._v.append(value)
        coeff_a = value[0]
        coeff_b = (value[1] - value[0])/(time[1] - time[0])
        self._c.append([coeff_a, coeff_b, 0.0, 0.0])
        # region 2 - linear rampup2
        time = _np.array([self._rampup2_start_time, self._rampdown_start_time])
        value = _np.array([self._rampup2_start_value,
                           self._rampdown_start_value])
        self._t.append(time)
        self._v.append(value)
        coeff_a = value[0]
        coeff_b = (value[1] - value[0])/(time[1] - time[0])
        self._c.append([coeff_a, coeff_b, 0.0, 0.0])
        # region 3 - linear rampdown
        time = _np.array([self._rampdown_start_time,
                          self._rampdown_stop_time])
        value = _np.array([self._rampdown_start_value,
                           self._rampdown_stop_value])
        self._t.append(time)
        self._v.append(value)
        coeff_a = value[0]
        coeff_b = (value[1] - value[0])/(time[1] - time[0])
        self._c.append([coeff_a, coeff_b, 0.0, 0.0])
        # region 4 - cubic
        self._t.append(_np.array([self._rampdown_stop_time,
                                  self._duration]))
        self._v.append(_np.array([self._rampdown_stop_value,
                                  self._start_value]))
        self._c.append([0.0, 0.0, 0.0, 0.0])

        # set coeffs for region 0
        idx = 0
        dt = self._t[idx][1] - self._t[idx][0]
        dv = self._v[idx][1] - self._v[idx][0]
        coeff_b_region1 = self._c[1][1]
        coeff_a = self._v[idx][0]
        coeff_b = 0.0
        coeff_c = (3*dv - coeff_b_region1*dt)/dt**2
        coeff_d = -(2*dv - coeff_b_region1*dt)/dt**3
        self._c[idx] = [coeff_a, coeff_b, coeff_c, coeff_d]

        # set coeffs for region 4
        idx = 4
        dt = self._t[idx][1] - self._t[idx][0]
        dv = self._v[idx][1] - self._v[idx][0]
        coeff_b_region3 = self._c[3][1]
        coeff_a = self._v[idx][0]
        coeff_b = coeff_b_region3
        coeff_c = (+3*(dv - coeff_b_region3*dt) + coeff_b_region3*dt)/dt**2
        coeff_d = (-2*(dv - coeff_b_region3*dt) - coeff_b_region3*dt)/dt**3
        self._c[idx] = [coeff_a, coeff_b, coeff_c, coeff_d]

        # define regions
        stime = self._rampup2_start_time
        # time1 = stime - (stime - self._rampup1_start_time)/4
        # time2 = stime + (self._rampdown_start_time - stime)/4
        time1 = stime - self._rampup_smooth_intvl/2
        time2 = stime + self._rampup_smooth_intvl/2
        self._region12_t = [time1, stime, time2]
        stime = self._rampdown_start_time
        # time1 = stime - (stime - self._rampup2_start_time)/4
        # time2 = stime + (self._rampdown_stop_time - stime)/4
        time1 = stime - self._rampdown_smooth_intvl/2
        time2 = stime + self._rampdown_smooth_intvl/2
        self._region23_t = [time1, stime, time2]

    def _func_region_12(self, time, delta=None):
        """Evaluate function in regions 1 and 2."""
        if delta is None:
            delta = self._rampup_smooth_value
        return self._func_region(time, delta, self._region12_t, 1, 2)

    def _func_region_23(self, time, delta=None):
        """Evaluate function in regions 2 and 3."""
        if delta is None:
            delta = self._rampdown_smooth_value
        return self._func_region(time, delta, self._region23_t, 2, 3)

    def _func_region(self, time, delta, region_t, region_idx1, region_idx2):
        """Evaluate function in regions (1 and 2) or (2 and 3)."""
        v = _np.zeros(time.shape)
        time1, stime, time2 = region_t
        sel = (time <= time1)
        v[sel] = self._func_polynom(region_idx1, time[sel])
        sel = (time >= time2)
        v[sel] = self._func_polynom(region_idx2, time[sel])
        sel = (time1 < time) & (time < time2)
        if _np.any(sel):
            poly1 = self._func_polynom(region_idx1, time[sel])
            poly2 = self._func_polynom(region_idx2, time[sel])
            vm = (poly1 + poly2) / 2
            vd = poly1 - poly2
            dt1, dt2 = stime - time1, time2 - stime
            T = dt1 + (dt2 - dt1)*(time[sel] - time1)/(time2 - time1)
            f3 = (time[sel] - stime)/T
            f4 = (1 + _np.cos(_np.pi*f3)) / 2
            vp = vm + 0.5 * _np.sqrt(vd**2 + (2*delta)**2*(f4**2)**0.5)
            vn = vm - 0.5 * _np.sqrt(vd**2 + (2*delta)**2*(f4**2)**0.5)
            b1 = self._c[region_idx1][1]
            b2 = self._c[region_idx2][1]
            vt = vp if b2 >= b1 else vn
            v[sel] = vt
        return v

    def _func_polynom(self, region_idx, time):
        """Return evaluation of polynomial in region of interest."""
        time = _np.tile(_np.reshape(time, (-1, 1)), len(WaveformParam.N))
        time0 = self._t[region_idx][0]
        dtime = time - time0
        dtpowern = dtime ** WaveformParam.N
        coeffs = self._c[region_idx]
        value = _np.dot(dtpowern, coeffs)
        return value


class _WaveformMagnet:
    """Base class of magnet waveforms."""

    def __init__(self, psname, wfm_nrpoints=_DEF_WFMSIZE_OTHERS):
        maname = _MASearch.conv_psname_2_psmaname(psname)
        self._magnet = _get_magnet(maname)
        self._psname = psname
        self._wfm_nrpoints = wfm_nrpoints

    @property
    def times(self):
        return _np.linspace(0, self.duration, self.wfm_nrpoints)

    @property
    def currents(self):
        return self._get_currents()

    @property
    def strengths(self):
        return self._get_strengths()

    @strengths.setter
    def strengths(self, value):
        self._set_strengths(value)

    @property
    def wfm_nrpoints(self):
        return self._wfm_nrpoints

    def __getitem__(self, index):
        """Return waveform at index."""
        wfm = self.waveform
        return wfm[index]

    def __iter__(self):
        """Return iterator for waveform."""
        wfm = self.waveform
        for i in range(len(wfm)):
            yield(wfm[i])

    def __reversed__(self):
        """Return reverse iterator for waveform."""
        wfm = self.waveform
        for i in range(len(wfm)-1, -1, -1):
            yield(wfm[i])

    def __len__(self):
        """Return length of waveform."""
        return self.wfm_nrpoints

    def __contains__(self, value):
        """Check whether value is in waveform."""
        return value in self.waveform

    def __eq__(self, value):
        """Compare waveforms."""
        return self.waveform == value

    def conv_current_2_strength(self, currents, **kwargs):
        return self._magnet.conv_current_2_strength(currents, **kwargs)

    def conv_strength_2_current(self, strengths, **kwargs):
        return self._magnet.conv_strength_2_current(strengths, **kwargs)

    def get_current_from_time(self, time):
        return _np.interp(time, self.times, self.currents)

    def get_strength_from_time(self, time):
        return _np.interp(time, self.times, self.strengths)


class WaveformDipole(_WaveformMagnet, WaveformParam):
    """Waveform for Dipole."""

    _E0 = _c.electron_rest_energy * _mu.joule_2_GeV

    def __init__(
            self, psname='BO-Fam:PS-B-1', wfm_nrpoints=_DEF_WFMSIZE_OTHERS,
            duration=_ru.DEFAULT_PS_RAMP_DURATION,
            start_energy=_ru.DEFAULT_PS_RAMP_START_ENERGY,
            rampup1_start_time=_ru.DEFAULT_PS_RAMP_RAMPUP1_START_TIME,
            rampup1_start_energy=_ru.DEFAULT_PS_RAMP_RAMPUP1_START_ENERGY,
            rampup2_start_time=_ru.DEFAULT_PS_RAMP_RAMPUP2_START_TIME,
            rampup2_start_energy=_ru.DEFAULT_PS_RAMP_RAMPUP2_START_ENERGY,
            rampup_smooth_intvl=_ru.DEFAULT_PS_RAMP_RAMPUP_SMOOTH_INTVL,
            rampup_smooth_energy=_ru.DEFAULT_PS_RAMP_RAMPUP_SMOOTH_ENERGY,
            rampdown_start_time=_ru.DEFAULT_PS_RAMP_RAMPDOWN_START_TIME,
            rampdown_start_energy=_ru.DEFAULT_PS_RAMP_RAMPDOWN_START_ENERGY,
            rampdown_stop_time=_ru.DEFAULT_PS_RAMP_RAMPDOWN_STOP_TIME,
            rampdown_stop_energy=_ru.DEFAULT_PS_RAMP_RAMPDOWN_STOP_ENERGY,
            rampdown_smooth_intvl=_ru.DEFAULT_PS_RAMP_RAMPDOWN_SMOOTH_INTVL,
            rampdown_smooth_energy=_ru.DEFAULT_PS_RAMP_RAMPDOWN_SMOOTH_ENERGY
            ):
        """Constructor."""
        _WaveformMagnet.__init__(self, psname, wfm_nrpoints)

        self._start_energy = start_energy
        self._rampup1_start_energy = rampup1_start_energy
        self._rampup2_start_energy = rampup2_start_energy
        self._rampdown_start_energy = rampdown_start_energy
        self._rampdown_stop_energy = rampdown_stop_energy
        self._rampup_smooth_energy = rampup_smooth_energy
        self._rampdown_smooth_energy = rampdown_smooth_energy

        _conv_func = self.conv_strength_2_current
        start_value = _conv_func(start_energy)
        rampup1_start_value = _conv_func(rampup1_start_energy)
        rampup2_start_value = _conv_func(rampup2_start_energy)
        rampdown_start_value = _conv_func(rampdown_start_energy)
        rampdown_stop_value = _conv_func(rampdown_stop_energy)
        rampup_smooth_value = _conv_func(rampup_smooth_energy)
        rampdown_smooth_value = _conv_func(rampdown_smooth_energy)
        WaveformParam.__init__(
            self,
            duration=duration,
            rampup1_start_time=rampup1_start_time,
            rampup2_start_time=rampup2_start_time,
            rampdown_start_time=rampdown_start_time,
            rampdown_stop_time=rampdown_stop_time,
            rampup_smooth_intvl=rampup_smooth_intvl,
            rampdown_smooth_intvl=rampdown_smooth_intvl,
            start_value=start_value,
            rampup1_start_value=rampup1_start_value,
            rampup2_start_value=rampup2_start_value,
            rampdown_start_value=rampdown_start_value,
            rampdown_stop_value=rampdown_stop_value,
            rampup_smooth_value=rampup_smooth_value,
            rampdown_smooth_value=rampdown_smooth_value)
        self._currents = self.eval_at(self.times)
        self._strengths = self.conv_current_2_strength(self._currents)

    @property
    def waveform(self):
        """Magnet waveform."""
        return self.strengths

    @property
    def start_energy(self):
        """Return start_energy."""
        return self._start_energy

    @property
    def rampup1_start_energy(self):
        """Return rampup1_start_energy."""
        return self._rampup1_start_energy

    @property
    def rampup2_start_energy(self):
        """Return rampup2_start_energy."""
        return self._rampup2_start_energy

    @property
    def rampdown_start_energy(self):
        """Return rampdown_start_energy."""
        return self._rampdown_start_energy

    @property
    def rampdown_stop_energy(self):
        """Return rampdown_stop_energy."""
        return self._rampdown_stop_energy

    @property
    def rampup_smooth_energy(self):
        """Return rampup_smooth_energy."""
        return self._rampup_smooth_energy

    @property
    def rampdown_smooth_energy(self):
        """Return rampdown_smooth_energy."""
        return self._rampdown_smooth_energy

    @start_energy.setter
    def start_energy(self, value):
        """Set start_energy."""
        value = float(value) if value > self._E0 else self._E0
        self._start_energy = float(value)
        current_value = self.conv_strength_2_current(value)
        WaveformParam.start_value.fset(self, current_value)

    @rampup1_start_energy.setter
    def rampup1_start_energy(self, value):
        """Set rampup1_start_energy."""
        value = float(value) if value > self._E0 else self._E0
        self._rampup1_start_energy = value
        current_value = self.conv_strength_2_current(value)
        WaveformParam.rampup1_start_value.fset(self, current_value)

    @rampup2_start_energy.setter
    def rampup2_start_energy(self, value):
        """Set rampup2_start_energy."""
        value = float(value) if value > self._E0 else self._E0
        self._rampup2_start_energy = value
        current_value = self.conv_strength_2_current(value)
        WaveformParam.rampup2_start_value.fset(self, current_value)

    @rampdown_start_energy.setter
    def rampdown_start_energy(self, value):
        """Set rampdown_start_energy."""
        value = float(value) if value > self._E0 else self._E0
        self._rampdown_start_energy = value
        current_value = self.conv_strength_2_current(value)
        WaveformParam.rampdown_start_value.fset(self, current_value)

    @rampdown_stop_energy.setter
    def rampdown_stop_energy(self, value):
        """Set rampdown_stop_energy."""
        value = float(value) if value > self._E0 else self._E0
        self._rampdown_stop_energy = value
        current_value = self.conv_strength_2_current(value)
        WaveformParam.rampdown_stop_value.fset(self, current_value)

    @rampup_smooth_energy.setter
    def rampup_smooth_energy(self, value):
        """Set rampup_smooth_energy."""
        self._rampup_smooth_energy = value
        current_value = self.conv_strength_2_current(value)
        WaveformParam.rampup_smooth_value.fset(self, current_value)

    @rampdown_smooth_energy.setter
    def rampdown_smooth_energy(self, value):
        """Set rampdown_smooth_energy."""
        self._rampdown_smooth_energy = value
        current_value = self.conv_strength_2_current(value)
        WaveformParam.rampdown_smooth_value.fset(self, current_value)

    @WaveformParam.start_value.setter
    def start_value(self, value):
        """Set start_value."""
        WaveformParam.start_value.fset(self, value)
        new_energy = self.conv_current_2_strength(float(value))
        if new_energy < self._E0:
            new_energy = self._E0
        self._start_energy = new_energy

    @WaveformParam.rampup1_start_value.setter
    def rampup1_start_value(self, value):
        """Set rampup1_start_value."""
        WaveformParam.rampup1_start_value.fset(self, value)
        new_energy = self.conv_current_2_strength(float(value))
        if new_energy < self._E0:
            new_energy = self._E0
        self._rampup1_start_energy = new_energy

    @WaveformParam.rampup2_start_value.setter
    def rampup2_start_value(self, value):
        """Set rampup2_start_value."""
        WaveformParam.rampup2_start_value.fset(self, value)
        new_energy = self.conv_current_2_strength(float(value))
        if new_energy < self._E0:
            new_energy = self._E0
        self._rampup2_start_energy = new_energy

    @WaveformParam.rampdown_start_value.setter
    def rampdown_start_value(self, value):
        """Set rampdown_start_value."""
        WaveformParam.rampdown_start_value.fset(self, value)
        new_energy = self.conv_current_2_strength(float(value))
        if new_energy < self._E0:
            new_energy = self._E0
        self._rampdown_start_energy = new_energy

    @WaveformParam.rampdown_stop_value.setter
    def rampdown_stop_value(self, value):
        """Set rampdown_stop_value."""
        WaveformParam.rampdown_stop_value.fset(self, value)
        new_energy = self.conv_current_2_strength(float(value))
        if new_energy < self._E0:
            new_energy = self._E0
        self._rampdown_stop_energy = new_energy

    @WaveformParam.rampup_smooth_value.setter
    def rampup_smooth_value(self, value):
        """Set rampup_smooth_value."""
        WaveformParam.rampup_smooth_value.fset(self, value)
        self._rampup_smooth_energy = self.conv_current_2_strength(value)

    @WaveformParam.rampdown_smooth_value.setter
    def rampdown_smooth_value(self, value):
        """Set rampdown_smooth_value."""
        WaveformParam.rampdown_smooth_value.fset(self, value)
        self._rampdown_smooth_energy = self.conv_current_2_strength(value)

    def _get_currents(self):
        self._currents = self.eval_at(self.times)
        return self._currents

    def _get_strengths(self):
        self._strengths = self.conv_current_2_strength(self.currents)
        return self._strengths

    def __str__(self):
        ppties = (
            'duration [ms]',
            'rampup1_start_time [ms]',
            'rampup2_start_time [ms]',
            'rampdown_start_time [ms]',
            'rampdown_stop_time [ms]',
            'rampup_smooth_intvl [ms]',
            'rampdown_smooth_intvl [ms]',
            'start_energy [GeV]',
            'rampup1_start_energy [GeV]',
            'rampup2_start_energy [GeV]',
            'rampdown_start_energy [GeV]',
            'rampdown_stop_energy [GeV]',
            'rampup_smooth_energy [GeV]',
            'rampdown_smooth_energy [GeV]',
            'start_value [A]',
            'rampup1_start_value [A]',
            'rampup2_start_value [A]',
            'rampdown_start_value [A]',
            'rampdown_stop_value [A]',
            'rampup_smooth_value [A]',
            'rampdown_smooth_value [A]',
        )
        maxlen = max(tuple(len(p) for p in ppties) + (len('name'),))
        strfmt = '{:<' + str(maxlen) + 's}: {}\n'
        text = ''
        for ppty in ppties:
            text += strfmt.format(ppty, getattr(self, ppty.split(' ')[0]))
        return text


class Waveform(_WaveformMagnet):
    """Waveform class for general magnets."""

    def __init__(self, psname, dipole=None, family=None, strengths=None,
                 currents=None, wfm_nrpoints=_DEF_WFMSIZE_OTHERS):
        """Constructor."""
        if dipole is None:
            raise ValueError('{} waveform needs an associated '
                             'dipole waveform!'.format(psname))
        _WaveformMagnet.__init__(self, psname, wfm_nrpoints=wfm_nrpoints)
        self._dipole = dipole
        self._family = family
        if currents is not None:
            strengths = self._conv_currents_2_strengths(currents)
        if strengths is None:
            if psname in _ru.NOMINAL_STRENGTHS:
                nom_strengths = _ru.NOMINAL_STRENGTHS[psname]
            else:
                nom_strengths = 0.0
            strengths = [nom_strengths, ] * self._wfm_nrpoints
        self._currents = self._conv_strengths_2_currents(strengths)
        self._strengths = self._conv_currents_2_strengths(self._currents)

    @property
    def duration(self):
        """Ramp duration."""
        return self._dipole.duration

    def update(self):
        """Update object."""
        if self._family is not None:
            self._family.update()
        self._currents = self._conv_strengths_2_currents(self._strengths)

    def _get_currents(self):
        self.update()
        return self._currents

    def _get_strengths(self):
        self.update()
        return self._strengths

    def _set_strengths(self, value):
        if isinstance(value, (int, float)):
            self._strengths = [value, ] * len(self._strengths)
        elif len(value) != len(self._strengths):
            raise ValueError('Incorrect length of passed strengths!')
        else:
            self._strengths = value.copy()

    def _conv_currents_2_strengths(self, currents):
        kwargs = {'currents': currents}
        if self._dipole is not None:
            if self._dipole.wfm_nrpoints != self._wfm_nrpoints:
                dip_strengths = _np.interp(
                    self.times, self._dipole.times, self._dipole.strengths)
            else:
                dip_strengths = self._dipole.strengths
            kwargs['strengths_dipole'] = dip_strengths
        if self._family is not None:
            if self._family.wfm_nrpoints != self._wfm_nrpoints:
                fam_strengths = _np.interp(
                    self.times, self._family.times, self._family.strengths)
            else:
                fam_strengths = self._family.strengths
            kwargs['strengths_family'] = fam_strengths
        return self.conv_current_2_strength(**kwargs)

    def _conv_strengths_2_currents(self, strengths):
        kwargs = {'strengths': strengths}
        if self._dipole is not None:
            if self._dipole.wfm_nrpoints != self._wfm_nrpoints:
                dip_strengths = _np.interp(
                    self.times, self._dipole.times, self._dipole.strengths)
            else:
                dip_strengths = self._dipole.strengths
            kwargs['strengths_dipole'] = dip_strengths
        if self._family is not None:
            if self._family.wfm_nrpoints != self._wfm_nrpoints:
                fam_strengths = _np.interp(
                    self.times, self._family.times, self._family.strengths)
            else:
                fam_strengths = self._family.strengths
            kwargs['strengths_family'] = fam_strengths
        return self.conv_strength_2_current(**kwargs)

    # --- list methods (strengths) ---

    def __getitem__(self, index):
        """Return waveform at index."""
        self.update()
        if isinstance(index, slice):
            wp = self._strengths[index]
            return _np.array([wp[i] for i in range(len(wp))])
        elif isinstance(index, int):
            return self._strengths[index]
        else:
            raise IndexError

    def __iter__(self):
        """Return iterator for waveform."""
        self.update()
        for i in range(len(self._strengths)):
            yield(self._strengths[i])

    def __reversed__(self):
        """Return reverse iterator for waveform."""
        self.update()
        for i in range(len(self._strengths)-1, -1, -1):
            yield(self._strengths[i])

    def __len__(self):
        """Return length of waveform."""
        return len(self._strengths)

    def __contains__(self, value):
        """Check whether value is in waveform."""
        return value in self._strengths

    def __eq__(self, value):
        """Compare waveforms."""
        return self._strengths == value



#!/usr/bin/env python-sirius

"""Unittest module for ll_database.py."""

from unittest import TestCase

from siriuspy import util
from siriuspy.timesys import csdev


PUB_INTERFACE = (
    'get_otp_database',
    'get_out_database',
    'get_afc_out_database',
    'get_evr_database',
    'get_eve_database',
    'get_afc_database',
    'get_fout_database',
    'get_event_database',
    'get_clock_database',
    'get_evg_database',
    'get_hl_trigger_database',
    'Const',
    'ETypes',
    )


class TestModule(TestCase):
    """Test module interface."""

    def test_public_interface(self):
        """Test module's public interface."""
        valid = util.check_public_interface_namespace(csdev, PUB_INTERFACE)
        self.assertTrue(valid)

    def test_get_otp_database(self):
        """Test get_otp_database."""
        # TODO: implement test!

    def test_get_out_database(self):
        """Test get_out_database."""
        # TODO: implement test!

    def test_get_afc_out_database(self):
        """Test get_afc_out_database."""
        # TODO: implement test!

    def test_get_evr_database(self):
        """Test get_evr_database."""
        # TODO: implement test!

    def test_get_eve_database(self):
        """Test get_eve_database."""
        # TODO: implement test!

    def test_get_afc_database(self):
        """Test get_afc_database."""
        # TODO: implement test!

    def test_get_fout_database(self):
        """Test get_fout_database."""
        # TODO: implement test!

    def test_get_event_database(self):
        """Test get_event_database."""
        # TODO: implement test!

    def test_get_clock_database(self):
        """Test get_clock_database."""
        # TODO: implement test!

    def test_get_evg_database(self):
        """Test get_evg_database."""
        # TODO: implement test!

    def test_get_hl_clock_database(self):
        """Test get_hl_clock_database."""
        # TODO: implement test!

    def test_get_hl_event_database(self):
        """Test get_hl_event_database."""
        # TODO: implement test!

    def test_get_hl_evg_database(self):
        """Test get_hl_evg_database."""
        # TODO: implement test!

    def test_get_hl_trigger_database(self):
        """Test get_hl_trigger_database."""
        # TODO: implement test!

    def test_events_hl2ll_map(self):
        """Test HL2LL_MAP."""
        # TODO: implement test!

    def test_events_ll2hl_map(self):
        """Test LL2HL_MAP."""
        # TODO: implement test!

    def test_events_ll_tmp(self):
        """Test LL_TMP."""
        # TODO: implement test!

    def test_events_hl_pref(self):
        """Test HL_PREF."""
        # TODO: implement test!

    def test_events_ll_codes(self):
        """Test LL_CODES."""
        # TODO: implement test!

    def test_events_ll_names(self):
        """Test LL_EVENTS."""
        # TODO: implement test!

    def test_events_modes(self):
        """Test MODES."""
        # TODO: implement test!

    def test_events_delay_types(self):
        """Test DELAY_TYPES."""
        # TODO: implement test!

    def test_clocks_states(self):
        """Test STATES."""
        # TODO: implement test!

    def test_clocks_ll_tmp(self):
        """Test LL_TMP."""
        # TODO: implement test!

    def test_clocks_hl_tmp(self):
        """Test HL_TMP."""
        # TODO: implement test!

    def test_clocks_hl_pref(self):
        """Test HL_PREF."""
        # TODO: implement test!

    def test_clocks_hl2ll_map(self):
        """Test HL2LL_MAP."""
        # TODO: implement test!

    def test_clocks_ll2hl_map(self):
        """Test LL2HL_MAP."""
        # TODO: implement test!

    def test_triggers_states(self):
        """Test STATES."""
        # TODO: implement test!

    def test_triggers_intlk(self):
        """Test INTLK."""
        # TODO: implement test!

    def test_triggers_polarities(self):
        """Test POLARITIES."""
        # TODO: implement test!

    def test_triggers_delay_types(self):
        """Test DELAY_TYPES."""
        # TODO: implement test!

    def test_triggers_src_ll(self):
        """Test SRC_LL."""
        # TODO: implement test!


class TestConst(TestCase):
    """Test Const class."""

    PUB_INTERFACE = (
        'AC_FREQUENCY',
        'RF_DIVISION',
        'RF_FREQUENCY',
        'BASE_FREQUENCY',
        'RF_PERIOD',
        'BASE_DELAY',
        'RF_DELAY',
        'FINE_DELAY',
        'EvtModes',
        'EvtDlyTyp',
        'ClockStates',
        'TrigStates',
        'TrigPol',
        'LowLvlLock',
        'TrigDlyTyp',
        'InInjTab',
        'TrigSrcLL',
        'EvtHL2LLMap',
        'EvtLL2HLMap',
        'EvtLL',
        'ClkHL2LLMap',
        'ClkLL2HLMap',
        'ClkLL',
        'HLTrigStatusLabels',
    )

    def test_public_interface(self):
        """Test module's public interface."""
        valid = util.check_public_interface_namespace(
            csdev.Const, TestConst.PUB_INTERFACE)
        self.assertTrue(valid)



"""Define properties of all timing devices and their connections."""

import re as _re
from copy import deepcopy as _dcopy
from threading import Lock as _Lock

from .. import clientweb as _web
from ..namesys import SiriusPVName as _PVName, Filter as _Filter

from .ps_search import PSSearch as _PSSearch

_TIMEOUT = 1.0


class LLTimeSearch:
    """Get the timing devices connections."""

    LLRegExp = _re.compile('([A-Z]+)([0-9]{0,2})', _re.IGNORECASE)

    # defines the relations between input and output of the timing devices.
    In2OutMap = {
        'EVG': {
            'UPLINK': (
                'OUT0', 'OUT1', 'OUT2', 'OUT3',
                'OUT4', 'OUT5', 'OUT6', 'OUT7',
                ),
            },
        'EVR': {
            'UPLINK': (
                'OTP0', 'OTP1', 'OTP2', 'OTP3', 'OTP4', 'OTP5',
                'OTP6', 'OTP7', 'OTP8', 'OTP9', 'OTP10', 'OTP11',
                'OTP12', 'OTP13', 'OTP14', 'OTP15', 'OTP16', 'OTP17',
                'OTP18', 'OTP19', 'OTP20', 'OTP21', 'OTP22', 'OTP23',
                'OUT0', 'OUT1', 'OUT2', 'OUT3',
                'OUT4', 'OUT5', 'OUT6', 'OUT7',
                ),
            },
        'EVE': {
            'UPLINK': (
                'OTP0', 'OTP1', 'OTP2', 'OTP3', 'OTP4', 'OTP5',
                'OTP6', 'OTP7', 'OTP8', 'OTP9', 'OTP10', 'OTP11',
                'OTP12', 'OTP13', 'OTP14', 'OTP15', 'OTP16', 'OTP17',
                'OTP18', 'OTP19', 'OTP20', 'OTP21', 'OTP22', 'OTP23',
                'OUT0', 'OUT1', 'OUT2', 'OUT3',
                'OUT4', 'OUT5', 'OUT6', 'OUT7',
                'RFOUT',
                ),
            },
        'AMCFPGAEVR': {
            'SFP8': (
                'FMC1CH1', 'FMC1CH2', 'FMC1CH3', 'FMC1CH4', 'FMC1CH5',
                'FMC2CH1', 'FMC2CH2', 'FMC2CH3', 'FMC2CH4', 'FMC2CH5',
                'CRT0', 'CRT1', 'CRT2', 'CRT3', 'CRT4',
                'CRT5', 'CRT6', 'CRT7',
                ),
            },
        'OEMultSFP': {
            'OE1': ('OUT1', ),
            'OE2': ('OUT2', ),
            'OE3': ('OUT3', ),
            'OE4': ('OUT4', ),
            },
        'OEMultPOF': {
            'IN1': ('OUT1', ),
            'IN2': ('OUT2', ),
            'IN3': ('OUT3', ),
            'IN4': ('OUT4', ),
            },
        'OESglPOF': {
            'IN': ('OUT', ),
            },
        'OESglSFP': {
            'INRX': ('OUT', 'INTX'),
            },
        'Fout': {
            'UPLINK': (
                'OUT0', 'OUT1', 'OUT2', 'OUT3',
                'OUT4', 'OUT5', 'OUT6', 'OUT7',
                ),
            },
        'UDC': {
            'SYNCIN': ('BCKPLN', 'SYNCOUT'),
            },
        'Crate': {
            'CRT0': ('CRT0', ),
            'CRT1': ('CRT1', ),
            'CRT2': ('CRT2', ),
            'CRT3': ('CRT3', ),
            'CRT4': ('CRT4', ),
            'CRT5': ('CRT5', ),
            'CRT6': ('CRT6', ),
            'CRT7': ('CRT7', ),
            },
        'OERFRx': {'OPTICALACP': ('SIGNAL', )},
        'OERFTx': {'SIGNAL': ('OPTICALACP', )},
        }
    In2OutMap['DIO'] = {
        'P{0:03d}'.format(i): ('P{0:03d}'.format(i), ) for i in range(110)}
    In2OutMap['DIO'].update({
        'P{0:02d}'.format(i): ('P{0:02d}'.format(i), ) for i in range(1, 25)})
    In2OutMap['DIO']['P052B'] = ('P052B', )
    In2OutMap['DIO']['P027B'] = ('P027B', )

    Out2InMap = dict()
    for dev, conns_ in In2OutMap.items():
        dic_ = dict()
        Out2InMap[dev] = dic_
        for conn1, conns in conns_.items():
            for conn2 in conns:
                dic_[conn2] = conn1
    del dev, conns_, dic_, conn1, conns, conn2  # cleanning class namespace.

    _conn_from_evg = dict()
    _conn_twds_evg = dict()
    _top_chain_devs = set()
    _final_receiver_devs = set()
    _all_devices = set()
    _trig_src_devs = set()
    _fout_devs = set()
    _evg_devs = set()
    _lock = _Lock()

    @classmethod
    def get_channel_input(cls, channel):
        """Get channel input method."""
        if not isinstance(channel, _PVName):
            channel = _PVName(channel)
        o2i = cls.Out2InMap.get(channel.dev)
        if o2i is None:
            return []
        conn = o2i.get(channel.propty)
        if conn is None:
            conn = cls.In2OutMap[channel.dev].get(channel.propty)
        if conn is None:
            return []
        elif isinstance(conn, str):
            conn = [conn, ]
        return [_PVName(channel.device_name + ':' + co) for co in conn]

    @classmethod
    def get_channel_output_port_pvname(cls, channel):
        if not isinstance(channel, _PVName):
            channel = _PVName(channel)
        outprt = channel.propty
        if outprt.startswith('FMC'):
            outprt = 'FMC' + outprt[3] + 'Ch{0:d}'.format(int(outprt[6])-1)
        elif outprt.startswith('CRT'):
            outprt = 'AMC' + outprt[3]
        elif outprt.startswith('OTP'):
            outprt = 'OTP{0:02d}'.format(int(outprt[3:]))
        return channel.substitute(propty=outprt)

    @classmethod
    def get_channel_internal_trigger_pvname(cls, channel):
        if not isinstance(channel, _PVName):
            channel = _PVName(channel)
        inttrig = channel.propty
        if inttrig.startswith('OUT'):
            inttrig = 'OTP{0:02d}'.format(12 + int(inttrig[-1]))
        elif inttrig.startswith(('FMC', 'CRT', 'OTP')):
            return cls.get_channel_output_port_pvname(channel)
        return channel.substitute(propty=inttrig)

    @classmethod
    def get_device_names(cls, filters=None, sorting=None):
        """
        Return a set with all devices of type type_dev.

        if type_dev is None, return all devices.
        """
        cls._get_timedata()
        return _Filter.process_filters(
                sorted(cls._all_devices), filters=filters, sorting=sorting)

    @classmethod
    def get_evg_name(cls):
        return cls.get_device_names({'dev': 'EVG'})[0]

    @classmethod
    def get_evg2fout_mapping(cls):
        fouts = cls.get_device_names({'dev': 'Fout'})
        link = list(cls.In2OutMap['Fout'])[0]
        mapp = dict()
        for fout in fouts:
            out = cls.get_evg_channel(fout.substitute(propty=link))
            mapp[out.propty] = fout.device_name
        return mapp

    @classmethod
    def get_triggersource_devices(cls):
        cls._get_timedata()
        return _dcopy(cls._trig_src_devs)

    @classmethod
    def get_trigsrc2fout_mapping(cls):
        cls._get_timedata()
        mapp = dict()
        for dev in cls._trig_src_devs:
            inp = list(cls.In2OutMap[dev.dev])[0]
            dev = dev.substitute(propty=inp)
            mapp[dev.device_name] = cls.get_fout_channel(dev)
        return mapp

    @classmethod
    def get_fout2trigsrc_mapping(cls):
        mapp = cls.get_trigsrc2fout_mapping()
        mapp2 = dict()
        for k, v in mapp.items():
            if v.device_name in mapp2:
                mapp2[v.device_name][v.propty] = k
            else:
                mapp2[v.device_name] = {v.propty: k}
        return mapp2

    @classmethod
    def get_device_tree(cls, channel):
        """Return a dictionary with the beaglebone to power supply mapping."""
        cls._get_timedata()
        twds_evg = cls._conn_twds_evg
        up_chan = _PVName(channel)
        if up_chan in twds_evg.keys():
            up_chan = list(twds_evg[up_chan])[0]
        up_channels = [up_chan]
        while up_chan.device_name not in cls._top_chain_devs:
            lst_chan = cls.get_channel_input(up_chan)
            if not lst_chan:
                raise Exception('Channel ', up_chan, ' not in dictionary.')
            up_chan = list(twds_evg[lst_chan[0]])[0]
            up_channels.append(up_chan)
        return up_channels

    @classmethod
    def reset(cls):
        """Reset data to initial value."""
        cls._conn_from_evg = dict()
        cls._conn_twds_evg = dict()
        cls._get_timedata()

    @staticmethod
    def server_online():
        """Return True/False if Sirius web server is online."""
        return _web.server_online()

    @classmethod
    def get_connections_from_evg(cls):
        """Return a dictionary with the beaglebone to power supply mapping."""
        cls._get_timedata()
        return _dcopy(cls._conn_from_evg)

    @classmethod
    def get_connections_twds_evg(cls):
        """Return a dictionary with the beaglebone to power supply mapping."""
        cls._get_timedata()
        return _dcopy(cls._conn_twds_evg)

    @classmethod
    def get_top_chain_senders(cls):
        """Return a dictionary with the beaglebone to power supply mapping."""
        cls._get_timedata()
        return _dcopy(cls._top_chain_devs)

    @classmethod
    def get_final_receivers(cls):
        """Return a dictionary with the beaglebone to power supply mapping."""
        cls._get_timedata()
        return _dcopy(cls._final_receiver_devs)

    @classmethod
    def get_relations_from_evg(cls):
        """Return a dictionary with the beaglebone to power supply mapping."""
        cls._get_timedata()
        return _dcopy(cls._devs_from_evg)

    @classmethod
    def get_relations_twds_evg(cls):
        """Return a dictionary with the beaglebone to power supply mapping."""
        cls._get_timedata()
        return _dcopy(cls._devs_twds_evg)

    @classmethod
    def has_clock(cls, ll_trigger):
        """Check whether a low level trigger has access to EVG clocks.

        Args:
            ll_trigger (SiriusPVName): Trigger name.

        Raises:
            Exception: When trigger receiver is not recognized.

        Returns:
            bool: True or False.

        """
        name = _PVName(ll_trigger)
        if name.dev == 'AMCFPGAEVR':
            return True
        elif name.dev in {'EVR', 'EVE'}:
            return name.propty.startswith('OUT')
        else:
            raise Exception('Error: ' + name)

    @classmethod
    def has_delay_type(cls, ll_trigger):
        """Check whether a low level trigger has delay type.

        Args:
            ll_trigger (SiriusPVName): Trigger name.

        Returns:
            bool: True or False.

        """
        name = _PVName(ll_trigger)
        return name.dev in {'EVR', 'EVE'} and name.propty.startswith('OUT')

    @classmethod
    def get_trigger_name(cls, channel):
        """Get name of the trigger associated with channel."""
        chan_tree = cls.get_device_tree(channel)
        for up_chan in chan_tree:
            if up_chan.device_name in cls._trig_src_devs:
                return up_chan

    @classmethod
    def get_fout_channel(cls, channel):
        """Get name of the FOUT channel associated with channel."""
        chan_tree = cls.get_device_tree(channel)
        for up_chan in chan_tree:
            if up_chan.device_name in cls._fout_devs:
                return up_chan

    @classmethod
    def get_evg_channel(cls, channel):
        """Get name of the EVG channel associated with channel."""
        chan_tree = cls.get_device_tree(channel)
        for up_chan in chan_tree:
            if up_chan.device_name in cls._evg_devs:
                return up_chan

    # --- private methods ---

    @classmethod
    def _add_entry_to_map(cls, which_map, conn, ele1, ele2):
        if which_map.lower().startswith('from'):
            mapp = cls._conn_from_evg
        else:
            mapp = cls._conn_twds_evg
        ele1 = _PVName(ele1+':'+conn)
        ele2 = _PVName(ele2+':'+conn)
        ele1_entry = mapp.get(ele1)
        if not ele1_entry:
            mapp[ele1] = {ele2}
        else:
            ele1_entry |= {ele2}

    @classmethod
    def _get_timedata(cls):
        with cls._lock:
            if cls._conn_from_evg:
                return
            if not _web.server_online():
                raise Exception('Could not connect with Consts Server!!')
            text = _web.timing_devices_mapping(timeout=_TIMEOUT)
            cls._parse_text_and_build_connection_mappings(text)
            cls._update_related_maps()

    # --- methods auxiliar to _get_timedata ---

    @classmethod
    def _parse_text_and_build_connection_mappings(cls, text):
        from_evg = dict()
        twds_evg = dict()
        lines = text.splitlines()
        for n, line in enumerate(lines, 1):
            line = line.strip()
            if not line or line[0] == '#':
                continue  # empty line
            out, inn = line.split()[:2]
            out, inn = _PVName(out), _PVName(inn)
            if out in from_evg.keys():
                from_evg[out] |= {inn}
            else:
                from_evg[out] = {inn}

            if inn in twds_evg.keys():
                print(
                    'Duplicate device input connection in line ' +
                    '{0:d}:\n\t {1:s}'.format(n, line))
                return
            else:
                twds_evg[inn] = {out}
        cls._conn_from_evg = from_evg
        cls._conn_twds_evg = twds_evg
        cls._add_udc_info()
        cls._add_crates_info()

    @classmethod
    def _add_crates_info(cls):
        """Add the information of Crate to BPMs to timing map."""
        conns = tuple(cls.In2OutMap['AMCFPGAEVR'].values())[0]
        conns = [v for v in conns if not v.startswith('FMC')]

        conn_dict = cls._get_crates_mapping()
        used = set()
        twds_evg = _dcopy(cls._conn_twds_evg)
        for chan in twds_evg.keys():
            bpms = conn_dict.get(chan.device_name)
            if bpms is None:
                continue
            used.add(chan.device_name)
            for bpm in bpms:
                for conn in conns:
                    cls._add_entry_to_map(
                        which_map='from', conn=conn,
                        ele1=chan.device_name, ele2=bpm)
                    cls._add_entry_to_map(
                        which_map='twds', conn=conn,
                        ele1=bpm, ele2=chan.device_name)
        # print(conn_dict.keys() - used)

    @classmethod
    def _get_crates_mapping(cls):
        data = _web.crates_mapping()
        crates = dict()
        mapping = dict()
        for line in data.splitlines():
            line = line.strip()
            if not line or line[0] == '#':
                continue  # empty line
            crate, dev, *_ = line.split()
            dev = _PVName(dev)
            if crate not in mapping and dev.dev == 'AMCFPGAEVR':
                crates[crate] = dev
                mapping[crates[crate]] = list()
            else:
                mapping[crates[crate]].append(dev)
        return mapping

    @classmethod
    def _add_udc_info(cls):
        """Add the information of bbb to PS to timing map."""
        data = _PSSearch.get_udc_dict()
        conn_dict = {udc: [x[0] for x in bsmps] for udc, bsmps in data.items()}
        conn = list(cls.In2OutMap['UDC'].values())[0][0]
        used = set()
        twds_evg = _dcopy(cls._conn_twds_evg)
        for chan in twds_evg.keys():
            pss = conn_dict.get(chan.device_name)
            if pss is None:
                continue
            used.add(chan.device_name)
            for ps in pss:
                cls._add_entry_to_map(
                    which_map='from', conn=conn,
                    ele1=chan.device_name, ele2=ps)
                cls._add_entry_to_map(
                    which_map='twds', conn=conn,
                    ele1=ps, ele2=chan.device_name)
        # print(conn_dict.keys() - used)

    @classmethod
    def _update_related_maps(cls):
        def _get_device_names(filters=None, sorting=None):
            return _Filter.process_filters(
                    sorted(cls._all_devices), filters=filters, sorting=sorting)

        cls._build_devices_relations()
        cls._top_chain_devs = (
            cls._devs_from_evg.keys() - cls._devs_twds_evg.keys())
        cls._final_receiver_devs = (
            cls._devs_twds_evg.keys() - cls._devs_from_evg.keys())
        cls._all_devices = (
            cls._devs_from_evg.keys() | cls._devs_twds_evg.keys())
        cls._trig_src_devs = set(
                _get_device_names({'dev': 'EVR'}) +
                _get_device_names({'dev': 'EVE'}) +
                _get_device_names({'dev': 'AMCFPGAEVR'}))
        cls._fout_devs = set(_get_device_names({'dev': 'Fout'}))
        cls._evg_devs = set(_get_device_names({'dev': 'EVG'}))

    @classmethod
    def _build_devices_relations(cls):
        simple_map = dict()
        for k, vs in cls._conn_from_evg.items():
            devs = {_PVName(v.device_name) for v in vs}
            devs |= simple_map.get(k.device_name, set())
            simple_map[_PVName(k.device_name)] = devs

        inv_map = dict()
        for k, vs in simple_map.items():
            for v in vs:
                devs = {k}
                devs |= inv_map.get(v, set())
                inv_map[v] = devs

        cls._devs_from_evg = simple_map
        cls._devs_twds_evg = inv_map



"""SI ID configuration."""
from copy import deepcopy as _dcopy


def get_dict():
    """Return configuration type dictionary."""
    module_name = __name__.split('.')[-1]
    _dict = {
        'config_type_name': module_name,
        'value': _dcopy(_template_dict),
        'check': False,
    }
    return _dict


# When using this type of configuration to set the machine,
# the list of PVs should be processed in the same order they are stored
# in the configuration. The second numeric parameter in the pair is the
# delay [s] the client should wait before setting the next PV.

_pvs_ids = [
    ['SI-06SB:ID-APU22:MaxPhaseSpeed-SP', 0.0, 0.0],  # [mm/s]
    ['SI-07SP:ID-APU22:MaxPhaseSpeed-SP', 0.0, 0.0],  # [mm/s]
    ['SI-08SB:ID-APU22:MaxPhaseSpeed-SP', 0.0, 0.0],  # [mm/s]
    ['SI-09SA:ID-APU22:MaxPhaseSpeed-SP', 0.0, 0.0],  # [mm/s]
    ['SI-11SP:ID-APU58:MaxPhaseSpeed-SP', 0.0, 0.0],  # [mm/s]
]

_template_dict = {
    'pvs': _pvs_ids,
}



"""ClientArchiver Exception Classes."""

import asyncio as _asyncio
import urllib as _urllib


class ClientArchError(Exception):
    """ClientArch Abstract Exception."""


class AuthenticationError(ClientArchError):
    """ClientArch Authentication Exception."""


class TimeoutError(ClientArchError, _asyncio.TimeoutError):
    """ClientArch Timeout Exception."""


class RuntimeError(ClientArchError, RuntimeError):
    """ClientArch Runtime Exception."""


class TypeError(ClientArchError, TypeError):
    """ClientArch TypeError Exception."""
    

class IndexError(ClientArchError, IndexError):
    """ClientArch IndexError Exception."""



"""Power Supply Diag App."""

from ... import csdev as _csdev
from ...namesys import SiriusPVName as _PVName
from ...search import PSSearch as _PSSearch


class ETypes(_csdev.ETypes):
    """Local enumerate types."""

    DIAG_STATUS_LABELS_AS = (
        'PS Disconnected/Comm. Broken',
        'PwrState-Sts Off',
        'Current-(SP|Mon) are different',
        'Interlocks',
        'Alarms',
        'OpMode-(Sel|Sts) are different',
        'Reserved')

    DIAG_STATUS_LABELS_LI = (
        'PS Disconnected/Comm. Broken',
        'PwrState-Sts Off',
        'Current-(SP|Mon) are different',
        'Interlocks',
        'Reserved',
        'Reserved',
        'Reserved')

    DIAG_STATUS_LABELS_FC = (
        'PS Disconnected',
        'PwrState-Sts Off',
        'Current-(SP|RB) are different',
        'Reserved',
        'Alarms',
        'OpMode-(Sel|Sts) are different',
        'Reserved')

    DIAG_STATUS_LABELS_BO = (
        'PS Disconnected/Comm. Broken',
        'PwrState-Sts Off',
        'Current-(SP|Mon) are different',
        'Interlocks',
        'Alarms',
        'OpMode-(Sel|Sts) are different',
        'Wfm error exceeded tolerance')


_et = ETypes  # syntactic sugar


def get_ps_diag_status_labels(psname):
    """Return Diag Status Labels enum."""
    psname = _PVName(psname)
    if psname.dev in ['FCH', 'FCV']:
        return _et.DIAG_STATUS_LABELS_FC
    if psname.sec == 'BO':
        return _et.DIAG_STATUS_LABELS_BO
    if psname.sec == 'LI':
        return _et.DIAG_STATUS_LABELS_LI
    return _et.DIAG_STATUS_LABELS_AS


def get_ps_diag_propty_database(psname):
    """Return property database of diagnostics for power supplies."""
    pstype = _PSSearch.conv_psname_2_pstype(psname)
    splims = _PSSearch.conv_pstype_2_splims(pstype)
    dtol = splims['DTOL_CUR']
    enums = get_ps_diag_status_labels(psname)
    dbase = {
        'DiagVersion-Cte': {'type': 'str', 'value': 'UNDEF'},
        'DiagCurrentDiff-Mon': {'type': 'float', 'value': 0.0,
                                'hilim': dtol, 'hihi': dtol, 'high': dtol,
                                'low': -dtol, 'lolo': -dtol, 'lolim': -dtol},
        'DiagStatus-Mon': {'type': 'int', 'value': 0,
                           'hilim': 1, 'hihi': 1, 'high': 1,
                           'low': -1, 'lolo': -1, 'lolim': -1
                           },
        'DiagStatusLabels-Cte': {'type': 'string', 'count': len(enums),
                                 'value': enums}
    }
    dbase = _csdev.add_pvslist_cte(dbase, 'Diag')
    return dbase



"""Definition of thread classes to used across the package."""

import time as _time
from threading import Thread as _Thread, Event as _Event, \
    Lock as _Lock
from queue import Queue as _Queue
from collections import deque as _deque


class AsyncWorker(_Thread):
    """Asynchronous Worker Thread.

    Performs asynchronous jobs indefinitely.
    """

    def __init__(self, name=None):
        """."""
        super().__init__(name=name, daemon=True)
        self._evt_received = _Event()
        self._evt_ready = _Event()
        self._evt_ready.set()
        self._evt_stop = _Event()
        self.target = None
        self.args = tuple()

    def configure_new_run(self, target, args=None):
        """Configure a new run of the thread."""
        if not self._evt_ready.is_set():
            return False
        self.target = target
        self.args = args or tuple()
        self._evt_ready.clear()
        # NOTE: _evt_received setting must be last operation of this method.
        self._evt_received.set()
        return True

    def wait_ready(self, timeout=None):
        """Wait until last run is finished."""
        self._evt_ready.wait(timeout=timeout)

    def is_ready(self):
        """Check if last run has finished."""
        return self._evt_ready.is_set()

    def stop(self):
        """Stop thread."""
        self._evt_stop.set()

    def run(self):
        """."""
        while not self._evt_stop.is_set():
            if self._evt_received.wait(0.5):
                self._evt_received.clear()
                self.target(*self.args)
                self._evt_ready.set()


class QueueThread(_Thread):
    """Callback queue class.

        Queues threads of this class are used to process callbacks of power
    supplies (magnets) properties among others.
    """

    # NOTE: QueueThread was reported as generating unstable behaviour
    # when used intensively in the SOFB IOC. Currently this class is
    # used in as-ps-diag IOC classes.
    # TODO: investigate this issue!

    def __init__(self):
        """Init method."""
        super().__init__(daemon=True)
        self._queue = _Queue()
        self._running = False

    @property
    def running(self):
        """Return whether thread is running."""
        return self._running

    def add_callback(self, func, *args, **kwargs):
        """Add callback."""
        if not hasattr(func, '__call__'):
            raise TypeError('Argument "func" is not callable.')
        self._queue.put((func, args, kwargs))

    def run(self):
        """Run method."""
        self._running = True
        while self.running:
            func_item = self._queue.get()
            nitems = self._queue.qsize()
            if nitems and nitems % 500 == 0:
                print("Warning: Queue size is {}!".format(nitems))
            function, args, kwargs = func_item
            function(*args, **kwargs)  # run the show!

    def stop(self):
        """Stop queue thread."""
        self._running = False


class RepeaterThread(_Thread):
    """Repeat execution of predefined function for a given number of times."""

    def __init__(self, interval, function, args=None, kwargs=None, niter=0):
        """Init method.

        Inputs:
        - interval: time to wait between sucessive calls to method.
        - function: method to be executed.
        - args: tuple of positional arguments of method
        - kwargs: dictionary with keyword arguments of method.
        - niter: number of times to execute method. If niter is zero or None
            it will be executed indefinetly until stop is called.
        """
        if args is None:
            args = tuple()
        if kwargs is None:
            kwargs = dict()
        super().__init__(daemon=True)
        self.interval = interval
        if not hasattr(function, '__call__'):
            raise TypeError('Argument "function" is not callable.')
        self.function = function
        self.args = args
        self.kwargs = kwargs
        self.niters = niter
        self.cur_iter = 0
        self._stopped = _Event()
        self._unpaused = _Event()
        self._unpaused.set()

    def run(self):
        """Run method."""
        self._unpaused.wait()
        self.function(*self.args, **self.kwargs)
        dtime = 0.0
        while ((not self._stopped.wait(self.interval - dtime)) and
               (not self.niters or self.niters > self.cur_iter)):
            self._unpaused.wait()
            if self._stopped.is_set():
                break
            self.cur_iter += 1
            _t0 = _time.time()
            self.function(*self.args, **self.kwargs)
            dtime = _time.time() - _t0

    def reset(self):
        """Reset count."""
        self.cur_iter = 0

    def pause(self):
        """."""
        self._unpaused.clear()

    def resume(self):
        """."""
        self._unpaused.set()

    def unpause(self):
        """."""
        self.resume()

    def is_paused(self):
        """."""
        return not self._unpaused.is_set()

    def isPaused(self):
        """."""
        return self.is_paused()

    def stop(self):
        """Stop execution."""
        self._unpaused.set()
        self._stopped.set()


class DequeThread(_deque):
    """DequeThread.

    This class manages generic operations (actions) using an append-right,
    pop-left queue. Each operation processing is a method invoked as a separate
    thread.
    """

    def __init__(self):
        """Init."""
        super().__init__()
        self._last_operation = None
        self._thread = None
        self._ignore = False
        self._enabled = True
        self._lock = _Lock()

    @property
    def last_operation(self):
        """Return last operation."""
        return self._last_operation

    @property
    def enabled(self):
        """Enable process."""
        return self._enabled

    @enabled.setter
    def enabled(self, value):
        self._enabled = value

    def ignore_set(self):
        """Turn ignore state on."""
        self._ignore = True

    def ignore_clear(self):
        """Turn ignore state on."""
        self._ignore = False

    def append(self, operation, unique=False):
        """Append operation to queue."""
        with self._lock:
            if self._ignore or (unique and self.count(operation) > 0):
                return False

            if not hasattr(operation, '__len__'):
                operation = (operation, )
            if not hasattr(operation[0], '__call__'):
                raise TypeError(
                    'First element of "operation" is not callable.')
            super().append(operation)
            self._last_operation = operation
            return True

    def clear(self):
        """Clear deque."""
        with self._lock:
            super().clear()

    def pop(self):
        """Pop operation from queue."""
        with self._lock:
            return super().pop()

    def popleft(self):
        """Pop left operation from queue."""
        with self._lock:
            return super().popleft()

    def process(self):
        """Process operation from queue."""
        # first check if a thread is already running
        with self._lock:
            donothing = not self._enabled
            donothing |= self._thread is not None and self._thread.is_alive()
            if donothing:
                return False

            # no thread is running, we can process queue
            try:
                operation = super().popleft()
            except IndexError:
                # there is nothing in the queue
                return False
            # process operation taken from queue
            args, kws = tuple(), dict()
            if len(operation) == 1:
                func = operation[0]
            elif len(operation) == 2:
                func, args = operation
            elif len(operation) >= 3:
                func, args, kws = operation[:3]
            self._thread = _Thread(
                target=func, args=args, kwargs=kws, daemon=True)
            self._thread.start()
            return True



"""."""
import time as _time
from threading import Event as _Flag
import numpy as _np

from .device import Device as _Device, Devices as _Devices, \
    ProptyDevice as _ProptyDevice
from ..diagbeam.bpm.csdev import Const as _csbpm
from ..search import BPMSearch as _BPMSearch
from ..namesys import SiriusPVName as _PVName


class BPM(_Device):
    """BPM Device."""

    _properties = (
        'asyn.ENBL', 'asyn.CNCT', 'SwMode-Sel', 'SwMode-Sts',
        'RFFEAtt-SP', 'RFFEAtt-RB',
        'SP_AArrayData', 'SP_BArrayData', 'SP_CArrayData', 'SP_DArrayData',
        'GEN_AArrayData', 'GEN_BArrayData', 'GEN_CArrayData', 'GEN_DArrayData',
        'GEN_XArrayData', 'GEN_YArrayData', 'GEN_SUMArrayData',
        'GEN_QArrayData',
        'GEN_RawXArrayData', 'GEN_RawYArrayData', 'GEN_RawSUMArrayData',
        'GEN_RawQArrayData',
        'SPPosX-Mon', 'SPPosY-Mon', 'SPSum-Mon', 'SPPosQ-Mon',
        'SPAmplA-Mon', 'SPAmplB-Mon', 'SPAmplC-Mon', 'SPAmplD-Mon',
        'PosX-Mon', 'PosY-Mon', 'Sum-Mon', 'PosQ-Mon',
        'AmplA-Mon', 'AmplB-Mon', 'AmplC-Mon', 'AmplD-Mon',
        'INFOClkFreq-RB', 'INFOHarmonicNumber-RB', 'INFOTBTRate-RB',
        'INFOFOFBRate-RB', 'INFOMONITRate-RB', 'INFOMONIT1Rate-RB',
        'GEN_PolyXArrayCoeff-SP', 'GEN_PolyXArrayCoeff-RB',
        'GEN_PolyYArrayCoeff-SP', 'GEN_PolyYArrayCoeff-RB',
        'GEN_PolySUMArrayCoeff-SP', 'GEN_PolySUMArrayCoeff-RB',
        'GEN_PolyQArrayCoeff-SP', 'GEN_PolyQArrayCoeff-RB',
        'PosKx-SP', 'PosKx-RB',
        'PosKy-RB', 'PosKy-SP',
        'PosKsum-SP', 'PosKsum-RB',
        'PosKq-SP', 'PosKq-RB',
        'PosXOffset-SP', 'PosXOffset-RB',
        'PosYOffset-SP', 'PosYOffset-RB',
        'PosSumOffset-SP', 'PosSumOffset-RB',
        'PosQOffset-SP', 'PosQOffset-RB',
        'ACQBPMMode-Sel', 'ACQBPMMode-Sts',
        'ACQChannel-Sel', 'ACQChannel-Sts',
        'ACQShots-SP', 'ACQShots-RB',
        'ACQUpdateTime-SP', 'ACQUpdateTime-RB',
        'ACQSamplesPre-SP', 'ACQSamplesPre-RB',
        'ACQSamplesPost-SP', 'ACQSamplesPost-RB',
        'ACQTriggerEvent-Sel', 'ACQTriggerEvent-Sts',
        'ACQStatus-Sts',
        'ACQTrigger-Sel', 'ACQTrigger-Sts',
        'ACQTriggerRep-Sel', 'ACQTriggerRep-Sts',
        'ACQDataTrigChan-Sel', 'ACQDataTrigChan-Sts',
        'ACQTriggerDataSel-SP', 'ACQTriggerDataSel-RB',
        'ACQTriggerDataThres-SP', 'ACQTriggerDataThres-RB',
        'ACQTriggerDataPol-Sel', 'ACQTriggerDataPol-Sts',
        'ACQTriggerDataHyst-SP', 'ACQTriggerDataHyst-RB',
        'SwTagEn-Sel', 'SwTagEn-Sts',
        'TbtTagEn-Sel', 'TbtTagEn-Sts',
        'Monit1TagEn-Sel', 'Monit1TagEn-Sts',
        'MonitTagEn-Sel', 'MonitTagEn-Sts',
        'TbtDataMaskEn-Sel', 'TbtDataMaskEn-Sts',
        'TbtDataMaskSamplesBeg-SP', 'TbtDataMaskSamplesBeg-RB',
        'TbtDataMaskSamplesEnd-SP', 'TbtDataMaskSamplesEnd-RB',
        'XYPosCal-Sel', 'XYPosCal-Sts',
        'SUMPosCal-Sel', 'SUMPosCal-Sts',
        'QPosCal-Sel', 'QPosCal-Sts',
        )

    CONV_NM2UM = 1e-3  # [nm] --> [um]

    def __init__(self, devname, auto_mon=True):
        """."""
        # call base class constructor
        if not _BPMSearch.is_valid_devname(devname):
            raise ValueError(devname + ' is not a valid BPM or PBPM name.')

        properties = set(BPM._properties)
        if _BPMSearch.is_photon_bpm(devname):
            properties -= {'RFFEAtt-SP', 'RFFEAtt-RB'}
        properties = list(properties)

        super().__init__(
            devname, properties=properties, auto_mon=auto_mon)
        self.csdata = _csbpm

    def __str__(self):
        """."""
        stg = '################### Summary Status ###################\n'
        stg += 'asyn:\n'
        stg += f'    Enabled: {_csbpm.EnblTyp._fields[self.asyn_state]:s}\n'
        stg += '    Connected: '
        stg += '    Switching Mode: '
        stg += f'{_csbpm.SwModes._fields[self.switching_mode]:s}\n'
        stg += f'{_csbpm.ConnTyp._fields[self.asyn_connected]:s}\n'
        stg += '\nAcquisition Parameters:\n'
        stg += f'    - Status: {_csbpm.AcqStates._fields[self.acq_status]:s}\n'
        stg += f'    - Mode: {_csbpm.OpModes._fields[self.acq_mode]:s}\n'
        stg += f'    - Channel: {_csbpm.AcqChan._fields[self.acq_channel]:s}\n'
        stg += f'    - Nr Shots: {self.acq_nrshots:d}\n'
        stg += f'    - Update Time: {self.acq_update_time:.1f} ms\n'
        stg += f'    - Repeat: {_csbpm.AcqRepeat._fields[self.acq_repeat]:s}\n'
        stg += '    - Trigger Type: '
        stg += f'{_csbpm.AcqTrigTyp._fields[self.acq_trigger]:s}\n'
        if self.acq_trigger == _csbpm.AcqTrigTyp.Data:
            stg += '        - Channel: '
            stg += f'{_csbpm.AcqChan._fields[self.acq_trig_datachan]:s}\n'
            stg += '        - Source: '
            stg += f'{_csbpm.AcqDataTyp._fields[self.acq_trig_datasel]:s}\n'
            stg += '        - Polarity: '
            stg += f'{_csbpm.Polarity._fields[self.acq_trig_datapol]:s}\n'
            stg += f'        - Threshold: {self.acq_trig_datathres:.1f}\n'
            stg += f'        - Hysteresis: {self.acq_trig_datahyst:d}\n'
        stg += '\n'
        return stg

    @property
    def is_ok(self):
        """."""
        stts = _csbpm.AcqStates
        okay = self.acq_status not in {
            stts.Error, stts.No_Memory, stts.Too_Few_Samples,
            stts.Too_Many_Samples, stts.Acq_Overflow}
        okay &= self.asyn_connected == _csbpm.ConnTyp.Connected
        okay &= self.asyn_state == _csbpm.EnblTyp.Enable
        return okay

    @property
    def rffe_att(self):
        """."""
        if 'RFFEAtt-RB' not in self._pvs:
            return None
        return self['RFFEAtt-RB']

    @rffe_att.setter
    def rffe_att(self, val):
        """."""
        if 'RFFEAtt-SP' in self._pvs:
            self['RFFEAtt-SP'] = val

    @property
    def asyn_state(self):
        """."""
        return self['asyn.ENBL']

    @asyn_state.setter
    def asyn_state(self, boo):
        """."""
        val = _csbpm.EnblTyp.Enable if boo else _csbpm.EnblTyp.Disable
        self['asyn.ENBL'] = val

    @property
    def asyn_connected(self):
        """."""
        return self['asyn.CNCT']

    @property
    def switching_mode(self):
        """."""
        return self['SwMode-Sts']

    @switching_mode.setter
    def switching_mode(self, val):
        """."""
        self['SwMode-Sel'] = val

    @property
    def adcfreq(self):
        """."""
        return self['INFOClkFreq-RB']

    @property
    def tbt_rate(self):
        """."""
        return self['INFOTBTRate-RB']

    @property
    def tbt_period(self):
        """."""
        return self.tbt_rate / self.adcfreq

    @property
    def tbt_sync_enbl(self):
        """."""
        return self['TbtTagEn-Sts']

    @tbt_sync_enbl.setter
    def tbt_sync_enbl(self, val):
        """."""
        self['TbtTagEn-Sel'] = val

    @property
    def tbt_mask_enbl(self):
        """."""
        return self['TbtDataMaskEn-Sts']

    @tbt_mask_enbl.setter
    def tbt_mask_enbl(self, val):
        """."""
        self['TbtDataMaskEn-Sel'] = val

    @property
    def tbt_mask_beg(self):
        """."""
        return self['TbtDataMaskSamplesBeg-RB']

    @tbt_mask_beg.setter
    def tbt_mask_beg(self, val):
        """."""
        self['TbtDataMaskSamplesBeg-SP'] = val

    @property
    def tbt_mask_end(self):
        """."""
        return self['TbtDataMaskSamplesEnd-RB']

    @tbt_mask_end.setter
    def tbt_mask_end(self, val):
        """."""
        self['TbtDataMaskSamplesEnd-SP'] = val

    @property
    def fofb_sync_enbl(self):
        """."""
        return self['SwTagEn-Sts']

    @fofb_sync_enbl.setter
    def fofb_sync_enbl(self, val):
        """."""
        self['SwTagEn-Sel'] = val

    @property
    def fofb_rate(self):
        """."""
        return self['INFOFOFBRate-RB']

    @property
    def fofb_period(self):
        """."""
        return self.fofb_rate / self.adcfreq

    @property
    def monit1_rate(self):
        """."""
        return self['INFOMONIT1Rate-RB']

    @property
    def monit1_period(self):
        """."""
        return self.monit1_rate / self.adcfreq

    @property
    def monit1_sync_enbl(self):
        """."""
        return self['Monit1TagEn']

    @monit1_sync_enbl.setter
    def monit1_sync_enbl(self, val):
        """."""
        self['Monit1TagEn-Sel'] = val

    @property
    def monit_rate(self):
        """."""
        return self['INFOMONITRate-RB']

    @property
    def monit_period(self):
        """."""
        return self.monit_rate / self.adcfreq

    @property
    def monit_sync_enbl(self):
        """."""
        return self['MonitTagEn-Sts']

    @monit_sync_enbl.setter
    def monit_sync_enbl(self, val):
        """."""
        self['MonitTagEn-Sel'] = val

    @property
    def posx_gain(self):
        """."""
        return self['PosKx-RB'] * self.CONV_NM2UM

    @posx_gain.setter
    def posx_gain(self, value):
        self['PosKx-SP'] = float(value) / self.CONV_NM2UM

    @property
    def posy_gain(self):
        """."""
        return self['PosKy-RB'] * self.CONV_NM2UM

    @posy_gain.setter
    def posy_gain(self, value):
        self['PosKy-SP'] = float(value) / self.CONV_NM2UM

    @property
    def possum_gain(self):
        """."""
        return self['PosKsum-RB']

    @possum_gain.setter
    def possum_gain(self, value):
        self['PosKsum-SP'] = float(value)

    @property
    def posq_gain(self):
        """."""
        return self['PosKq-RB']

    @posq_gain.setter
    def posq_gain(self, value):
        self['PosKq-SP'] = float(value)

    @property
    def posx_offset(self):
        """."""
        return self['PosXOffset-RB'] * self.CONV_NM2UM

    @posx_offset.setter
    def posx_offset(self, value):
        self['PosXOffset-SP'] = float(value) / self.CONV_NM2UM

    @property
    def posy_offset(self):
        """."""
        return self['PosYOffset-RB'] * self.CONV_NM2UM

    @posy_offset.setter
    def posy_offset(self, value):
        self['PosYOffset-SP'] = float(value) / self.CONV_NM2UM

    @property
    def possum_offset(self):
        """."""
        return self['PosSumOffset-RB']

    @possum_offset.setter
    def possum_offset(self, value):
        self['PosSumOffset-SP'] = float(value)

    @property
    def posq_offset(self):
        """."""
        return self['PosQOffset-RB']

    @posq_offset.setter
    def posq_offset(self, value):
        self['PosQOffset-SP'] = float(value)

    @property
    def posx(self):
        """."""
        return self['PosX-Mon'] * self.CONV_NM2UM

    @property
    def posy(self):
        """."""
        return self['PosY-Mon'] * self.CONV_NM2UM

    @property
    def possum(self):
        """."""
        return self['Sum-Mon']

    @property
    def posq(self):
        """."""
        return self['PosQ-Mon']

    @property
    def ampla(self):
        """."""
        return self['AmplA-Mon']

    @property
    def amplb(self):
        """."""
        return self['AmplB-Mon']

    @property
    def amplc(self):
        """."""
        return self['AmplC-Mon']

    @property
    def ampld(self):
        """."""
        return self['AmplD-Mon']

    @property
    def mt_posx(self):
        """."""
        return self['GEN_XArrayData'] * self.CONV_NM2UM

    @property
    def mt_posy(self):
        """."""
        return self['GEN_YArrayData'] * self.CONV_NM2UM

    @property
    def mt_possum(self):
        """."""
        return self['GEN_SUMArrayData']

    @property
    def mt_posq(self):
        """."""
        return self['GEN_QArrayData']

    @property
    def mt_ampla(self):
        """."""
        return self['GEN_AArrayData']

    @property
    def mt_amplb(self):
        """."""
        return self['GEN_BArrayData']

    @property
    def mt_amplc(self):
        """."""
        return self['GEN_CArrayData']

    @property
    def mt_ampld(self):
        """."""
        return self['GEN_DArrayData']

    @property
    def mt_polyx(self):
        """."""
        return self['GEN_PolyXArrayCoeff-RB']

    @mt_polyx.setter
    def mt_polyx(self, value):
        """."""
        self['GEN_PolyXArrayCoeff-SP'] = _np.array(value)

    @property
    def mt_polyy(self):
        """."""
        return self['GEN_PolyYArrayCoeff-RB']

    @mt_polyy.setter
    def mt_polyy(self, value):
        """."""
        self['GEN_PolyYArrayCoeff-SP'] = _np.array(value)

    @property
    def mt_polysum(self):
        """."""
        return self['GEN_PolySUMArrayCoeff-RB']

    @mt_polysum.setter
    def mt_polysum(self, value):
        """."""
        self['GEN_PolySUMArrayCoeff-SP'] = _np.array(value)

    @property
    def mt_polyq(self):
        """."""
        return self['GEN_PolyQArrayCoeff-RB']

    @mt_polyq.setter
    def mt_polyq(self, value):
        """."""
        self['GEN_PolyQArrayCoeff-SP'] = _np.array(value)

    @property
    def mt_polyxy_enbl(self):
        """."""
        return self['XYPosCal-Sts']

    @mt_polyxy_enbl.setter
    def mt_polyxy_enbl(self, val):
        """."""
        self['XYPosCal-Sel'] = val

    @property
    def mt_polysum_enbl(self):
        """."""
        return self['SUMPosCal-Sts']

    @mt_polysum_enbl.setter
    def mt_polysum_enbl(self, val):
        """."""
        self['SUMPosCal-Sel'] = val

    @property
    def mt_polyq_enbl(self):
        """."""
        return self['QPosCal-Sts']

    @mt_polyq_enbl.setter
    def mt_polyq_enbl(self, val):
        """."""
        self['QPosCal-Sel'] = val

    @property
    def mtraw_posx(self):
        """Multi turn raw X array data."""
        return self['GEN_RawXArrayData']

    @property
    def mtraw_posy(self):
        """Multi turn raw Y array data."""
        return self['GEN_RawYArrayData']

    @property
    def mtraw_possum(self):
        """Multi turn raw sum array data."""
        return self['GEN_RawSUMArrayData']

    @property
    def mtraw_posq(self):
        """Multi turn raw Q array data."""
        return self['GEN_RawQArrayData']

    @property
    def sp_posx(self):
        """."""
        return self['SPPosX-Mon'] * self.CONV_NM2UM

    @property
    def sp_posy(self):
        """."""
        return self['SPPosY-Mon'] * self.CONV_NM2UM

    @property
    def sp_possum(self):
        """."""
        return self['SPSum-Mon']

    @property
    def sp_posq(self):
        """."""
        return self['SPPosQ-Mon']

    @property
    def sp_ampla(self):
        """."""
        return self['SPAmplA-Mon']

    @property
    def sp_amplb(self):
        """."""
        return self['SPAmplB-Mon']

    @property
    def sp_amplc(self):
        """."""
        return self['SPAmplC-Mon']

    @property
    def sp_ampld(self):
        """."""
        return self['SPAmplD-Mon']

    @property
    def sp_arraya(self):
        """."""
        return self['SP_AArrayData']

    @property
    def sp_arrayb(self):
        """."""
        return self['SP_BArrayData']

    @property
    def sp_arrayc(self):
        """."""
        return self['SP_CArrayData']

    @property
    def sp_arrayd(self):
        """."""
        return self['SP_DArrayData']

    @property
    def acq_mode(self):
        """."""
        return self['ACQBPMMode-Sts']

    @acq_mode.setter
    def acq_mode(self, mode):
        """."""
        self['ACQBPMMode-Sel'] = mode

    @property
    def acq_ctrl(self):
        """."""
        return self['ACQTriggerEvent-Sts']

    @acq_ctrl.setter
    def acq_ctrl(self, val):
        """."""
        self['ACQTriggerEvent-Sel'] = val

    @property
    def acq_status(self):
        """."""
        return self['ACQStatus-Sts']

    @property
    def acq_channel(self):
        """."""
        return self['ACQChannel-Sts']

    @acq_channel.setter
    def acq_channel(self, val):
        """."""
        self['ACQChannel-Sel'] = val

    @property
    def acq_trigger(self):
        """."""
        return self['ACQTrigger-Sts']

    @acq_trigger.setter
    def acq_trigger(self, val):
        """."""
        self['ACQTrigger-Sel'] = val

    @property
    def acq_repeat(self):
        """."""
        return self['ACQTriggerRep-Sts']

    @acq_repeat.setter
    def acq_repeat(self, val):
        """."""
        self['ACQTriggerRep-Sel'] = val

    @property
    def acq_update_time(self):
        """."""
        return self['ACQUpdateTime-RB'] / 1e3

    @acq_update_time.setter
    def acq_update_time(self, val):
        """."""
        self['ACQUpdateTime-SP'] = val * 1e3

    @property
    def acq_trig_datachan(self):
        """."""
        return self['ACQDataTrigChan-Sts']

    @acq_trig_datachan.setter
    def acq_trig_datachan(self, val):
        """."""
        self['ACQDataTrigChan-Sel'] = val

    @property
    def acq_trig_datasel(self):
        """."""
        return self['ACQTriggerDataSel-RB']

    @acq_trig_datasel.setter
    def acq_trig_datasel(self, val):
        """."""
        self['ACQTriggerDataSel-SP'] = val

    @property
    def acq_trig_datathres(self):
        """."""
        return self['ACQTriggerDataThres-RB']

    @acq_trig_datathres.setter
    def acq_trig_datathres(self, val):
        """."""
        self['ACQTriggerDataThres-SP'] = val

    @property
    def acq_trig_datahyst(self):
        """."""
        return self['ACQTriggerDataHyst-RB']

    @acq_trig_datahyst.setter
    def acq_trig_datahyst(self, val):
        """."""
        self['ACQTriggerDataHyst-SP'] = val

    @property
    def acq_trig_datapol(self):
        """."""
        return self['ACQTriggerDataPol-RB']

    @acq_trig_datapol.setter
    def acq_trig_datapol(self, val):
        """."""
        self['ACQTriggerDataPol-SP'] = val

    @property
    def acq_nrsamples_post(self):
        """."""
        return self['ACQSamplesPost-RB']

    @acq_nrsamples_post.setter
    def acq_nrsamples_post(self, val):
        """."""
        self['ACQSamplesPost-SP'] = val

    @property
    def acq_nrsamples_pre(self):
        """."""
        return self['ACQSamplesPre-RB']

    @acq_nrsamples_pre.setter
    def acq_nrsamples_pre(self, val):
        """."""
        self['ACQSamplesPre-SP'] = val

    @property
    def acq_nrshots(self):
        """."""
        return self['ACQShots-RB']

    @acq_nrshots.setter
    def acq_nrshots(self, val):
        """."""
        self['ACQShots-SP'] = val

    def wait_acq_finish(self, timeout=10):
        """Wait Acquisition to finish."""
        vals = {
            _csbpm.AcqStates.Idle, _csbpm.AcqStates.Error,
            _csbpm.AcqStates.Aborted, _csbpm.AcqStates.Too_Many_Samples,
            _csbpm.AcqStates.Too_Few_Samples, _csbpm.AcqStates.No_Memory,
            _csbpm.AcqStates.Acq_Overflow}
        return self._wait(
            'ACQStatus-Sts', vals, timeout=timeout, comp=lambda x, y: x in y)

    def wait_acq_start(self, timeout=10):
        """Wait Acquisition to start."""
        vals = {
            _csbpm.AcqStates.Waiting, _csbpm.AcqStates.External_Trig,
            _csbpm.AcqStates.Data_Trig, _csbpm.AcqStates.Software_Trig,
            _csbpm.AcqStates.Acquiring}
        return self._wait(
            'ACQStatus-Sts', vals, timeout=timeout, comp=lambda x, y: x in y)

    def cmd_acq_start(self):
        """Command Start Acquisition."""
        self.acq_ctrl = _csbpm.AcqEvents.Start
        return self._wait('ACQTriggerEvent-Sts', _csbpm.AcqEvents.Start)

    def cmd_acq_stop(self):
        """Command Stop Acquisition."""
        self.acq_ctrl = _csbpm.AcqEvents.Stop
        return self._wait('ACQTriggerEvent-Sts', _csbpm.AcqEvents.Stop)

    def cmd_acq_abort(self):
        """Command Abort Acquisition."""
        self.acq_ctrl = _csbpm.AcqEvents.Abort
        return self._wait('ACQTriggerEvent-Sts', _csbpm.AcqEvents.Abort)

    def cmd_turn_on_switching(self):
        """Command Turn on Switching."""
        self.switching_mode = _csbpm.SwModes.switching
        return self._wait('SwMode-Sts', _csbpm.SwModes.switching)

    def cmd_turn_off_switching(self):
        """Command Turn off Switching."""
        self.switching_mode = _csbpm.SwModes.direct
        return self._wait('SwMode-Sts', _csbpm.SwModes.direct)

    def cmd_sync_tbt(self):
        """Synchronize TbT acquisitions with Timing System."""
        self.tbt_sync_enbl = 1
        _time.sleep(0.1)
        self.tbt_sync_enbl = 0
        return self._wait('TbtTagEn-Sts', 0)

    def cmd_sync_fofb(self):
        """Synchronize FOFB acquisitions with Timing System."""
        self.fofb_sync_enbl = 1
        _time.sleep(0.1)
        self.fofb_sync_enbl = 0
        return self._wait('SwTagEn-Sts', 0)

    def cmd_sync_monit1(self):
        """Synchronize Monit1 acquisitions with Timing System."""
        self.monit1_sync_enbl = 1
        _time.sleep(0.1)
        self.monit1_sync_enbl = 0
        return self._wait('Monit1TagEn-Sts', 0)

    def cmd_sync_monit(self):
        """Synchronize Monit acquisitions with Timing System."""
        self.monit_sync_enbl = 1
        _time.sleep(0.1)
        self.monit_sync_enbl = 0
        return self._wait('Monit1TagEn-Sts', 0)


class FamBPMs(_Devices):
    """."""

    TIMEOUT = 10
    RFFEATT_MAX = 30

    class DEVICES:
        """."""
        SI = 'SI-Fam:DI-BPM'
        BO = 'BO-Fam:DI-BPM'
        ALL = (BO, SI)

    def __init__(self, devname=None):
        """."""
        if devname is None:
            devname = self.DEVICES.SI
        if devname not in self.DEVICES.ALL:
            raise ValueError('Wrong value for devname')

        devname = _PVName(devname)
        bpm_names = _BPMSearch.get_names(
            filters={'sec': devname.sec, 'dev': devname.dev})
        devs = [BPM(dev, auto_mon=False) for dev in bpm_names]

        super().__init__(devname, devs)
        self._bpm_names = bpm_names
        self._csbpm = devs[0].csdata
        propties_to_keep = ['GEN_XArrayData', 'GEN_YArrayData']

        self._mturn_flags = dict()
        for bpm in devs:
            for propty in propties_to_keep:
                bpm.set_auto_monitor(propty, True)
                pvo = bpm.pv_object(propty)
                self._mturn_flags[pvo.pvname] = _Flag()
                pvo.add_callback(self._mturn_set_flag)

    @property
    def bpm_names(self):
        """Return BPM names."""
        return self._bpm_names

    @property
    def csbpm(self):
        """Return control system BPM constants class."""
        return self._csbpm

    def set_attenuation(self, value=RFFEATT_MAX, timeout=TIMEOUT):
        """."""
        for bpm in self:
            bpm.rffe_att = value

        mstr = ''
        okall = True
        t0 = _time.time()
        for bpm in self:
            tout = timeout - (_time.time() - t0)
            if not bpm._wait('RFFEAtt-RB', value, timeout=tout):
                okall = False
                mstr += (
                    f'\n{bpm.devname:<20s}: ' +
                    f'rb {bpm.rffe_att:.0f} != sp {value:.0f}')

        print('RFFE attenuation set confirmed in all BPMs', end='')
        print(', except:' + mstr if mstr else '.')
        return okall

    def get_slow_orbit(self):
        """Get slow orbit vectors.

        Returns:
            orbx (numpy.ndarray, 160): Horizontal Orbit.
            orby (numpy.ndarray, 160): Vertical Orbit.

        """
        orbx, orby = [], []
        for bpm in self._devices:
            orbx.append(bpm.posx)
            orby.append(bpm.posy)
        orbx = _np.array(orbx)
        orby = _np.array(orby)
        return orbx, orby

    def get_mturn_orbit(self, return_sum=False):
        """Get Multiturn orbit matrices.

        Args:
            return_sum (bool, optional): Whether or not to return BPMs sum.
                Defaults to False.

        Returns:
            orbx (numpy.ndarray, Nx160): Horizontal Orbit.
            orby (numpy.ndarray, Nx160): Vertical Orbit.
            possum (numpy.ndarray, Nx160): BPMs Sum signal.

        """
        orbx, orby = [], []
        if return_sum:
            possum = []

        mini = None
        for bpm in self._devices:
            mtx = bpm.mt_posx
            mty = bpm.mt_posy
            orbx.append(mtx)
            orby.append(mty)

            if mini is None:
                mini = mtx.size
            mini = _np.min([mini, mtx.size, mty.size])

            if return_sum:
                mts = bpm.mt_possum
                possum.append(mts)
                mini = min(mini, mts.size)

        for i, (obx, oby) in enumerate(zip(orbx, orby)):
            orbx[i] = obx[:mini]
            orby[i] = oby[:mini]
            if return_sum:
                possum[i] = possum[i][:mini]
        orbx = _np.array(orbx).T
        orby = _np.array(orby).T

        if not return_sum:
            return orbx, orby
        return orbx, orby, _np.array(possum).T

    @staticmethod
    def get_sampling_frequency(rf_freq: float, acq_rate='Monit1'):
        """Return the sampling frequency of the acquisition.

        Args:
            rf_freq (float): RF frequency.
            acq_rate (str, optional): acquisition rate. Defaults to 'Monit1'.

        Returns:
            float: acquisition frequency.

        """
        fsamp = rf_freq / 864
        if acq_rate.lower().startswith('tbt'):
            return fsamp
        fsamp /= 23
        if acq_rate.lower().startswith('fofb'):
            return fsamp
        fsamp /= 25
        return fsamp

    def mturn_config_acquisition(
            self, nr_points_after: int, nr_points_before=0,
            acq_rate='Monit1', repeat=True, external=True):
        """Configure acquisition for BPMs.

        Args:
            nr_points_after (int): number of points after trigger.
            nr_points_before (int): number of points after trigger.
                Defaults to 0.
            acq_rate (str, optional): Acquisition rate ('TbT', 'FOFB',
                'Monit1'). Defaults to 'Monit1'.
            repeat (bool, optional): Whether or not acquisition should be
                repetitive. Defaults to True.
            external (bool, optional): Whether or not external trigger should
                be used. Defaults to True.

        """
        if acq_rate.lower().startswith('monit1'):
            acq_rate = self._csbpm.AcqChan.Monit1
        elif acq_rate.lower().startswith('fofb'):
            acq_rate = self._csbpm.AcqChan.FOFB
        elif acq_rate.lower().startswith('tbt'):
            acq_rate = self._csbpm.AcqChan.TbT
        else:
            raise ValueError(acq_rate + ' is not a valid acquisition rate.')

        if repeat:
            repeat = self._csbpm.AcqRepeat.Repetitive
        else:
            repeat = self._csbpm.AcqRepeat.Normal

        if external:
            trig = self._csbpm.AcqTrigTyp.External
        else:
            trig = self._csbpm.AcqTrigTyp.Now

        self.cmd_mturn_acq_abort()

        for bpm in self._devices:
            bpm.acq_repeat = repeat
            bpm.acq_channel = acq_rate
            bpm.acq_trigger = trig
            bpm.acq_nrsamples_pre = nr_points_before
            bpm.acq_nrsamples_post = nr_points_after

        self.cmd_mturn_acq_start()

    def cmd_mturn_acq_abort(self) -> bool:
        """Abort BPMs acquistion.

        Returns:
            bool: Whether or not abort was successful.

        """
        for bpm in self._devices:
            bpm.acq_ctrl = self._csbpm.AcqEvents.Abort

        for bpm in self._devices:
            boo = bpm.wait_acq_finish()
            if not boo:
                return False
        return True

    def cmd_mturn_acq_start(self) -> bool:
        """Start BPMs acquisition.

        Returns:
            bool: Whether or not start was successful.

        """
        for bpm in self._devices:
            bpm.acq_ctrl = self._csbpm.AcqEvents.Start

        for bpm in self._devices:
            boo = bpm.wait_acq_start()
            if not boo:
                return False
        return True

    def set_switching_mode(self, mode='direct'):
        """Set switching mode of BPMS.

        Args:
            mode ((str, int), optional): Desired mode, must be in
                {'direct', 'switching', 1, 3}. Defaults to 'direct'.

        Raises:
            ValueError: When mode is not in {'direct', 'switching', 1, 3}.

        """
        if mode not in ('direct', 'switching', 1, 3):
            raise ValueError('Value must be in ("direct", "switching", 1, 3).')

        for bpm in self._devices:
            bpm.switching_mode = mode

    def mturn_reset_flags(self):
        """Reset Multiturn flags to wait for a new orbit update."""
        for flag in self._mturn_flags.values():
            flag.clear()

    def mturn_wait_update_flags(self, timeout=10) -> int:
        """Wait Multiturn orbit update.

        Args:
            timeout (int, optional): Waiting timeout. Defaults to 10.

        Returns:
            int: code describing what happened:
                -2: TypeError ocurred
                -1: Size of X orbit is different than Y orbit.
                0: Orbit updated.
                >0: Index of the first BPM which did not updated plus 1.

        """
        orbx0, orby0 = self.get_mturn_orbit()
        for i, flag in enumerate(self._mturn_flags.values()):
            t00 = _time.time()
            if not flag.wait(timeout=timeout):
                return (i // 2) + 1
            timeout -= _time.time() - t00
            timeout = max(timeout, 0)

        while timeout > 0:
            t00 = _time.time()
            orbx, orby = self.get_mturn_orbit()
            typ = False
            try:
                sizx = min(orbx.shape[0], orbx0.shape[0])
                sizy = min(orby.shape[0], orby0.shape[0])
                continue_ = sizx != sizy
                errx = _np.all(
                    _np.isclose(orbx0[:sizx], orbx[:sizx]), axis=0)
                erry = _np.all(
                    _np.isclose(orby0[:sizy], orby[:sizy]), axis=0)
                continue_ |= _np.any(errx) | _np.any(erry)
            except TypeError:
                print('TypeError')
                typ = True
                continue_ = True
            if not continue_:
                return 0
            _time.sleep(0.1)
            timeout -= _time.time() - t00

        if typ:
            return -1
        elif sizx != sizy:
            return -2
        elif _np.any(errx):
            return int(errx.nonzero()[0][0])+1
        elif _np.any(erry):
            return int(erry.nonzero()[0][0])+1
        return False

    def _mturn_set_flag(self, pvname, **kwargs):
        _ = kwargs
        self._mturn_flags[pvname].set()


class BPMLogicalTrigger(_ProptyDevice):
    """BPM Logical Trigger device."""

    _properties = (
        'RcvSrc-Sel', 'RcvSrc-Sts',
        'RcvInSel-SP', 'RcvInSel-RB',
        'TrnSrc-Sel', 'TrnSrc-Sts',
        'TrnOutSel-SP', 'TrnOutSel-RB',
    )

    def __init__(self, bpmname, index):
        """Init."""
        if not _BPMSearch.is_valid_devname(bpmname):
            raise NotImplementedError(bpmname)
        if not 0 <= int(index) <= 23:
            raise NotImplementedError(index)
        super().__init__(
            bpmname, 'TRIGGER'+str(index),
            properties=BPMLogicalTrigger._properties)

    @property
    def receiver_source(self):
        """Receiver source."""
        return self['RcvSrc-Sts']

    @receiver_source.setter
    def receiver_source(self, value):
        self['RcvSrc-Sel'] = value

    @property
    def receiver_in_sel(self):
        """Receiver in selection."""
        return self['RcvInSel-RB']

    @receiver_in_sel.setter
    def receiver_in_sel(self, value):
        self['RcvInSel-SP'] = value

    @property
    def transmitter_source(self):
        """Transmitter source."""
        return self['TrnSrc-Sts']

    @transmitter_source.setter
    def transmitter_source(self, value):
        self['TrnSrc-Sel'] = value

    @property
    def transmitter_out_sel(self):
        """Transmitter out selection."""
        return self['TrnOutSel-RB']

    @transmitter_out_sel.setter
    def transmitter_out_sel(self, value):
        self['TrnOutSel-SP'] = value



"""Simulators subpackage."""

DBASE_DEFAULT = {'type': 'float', 'value': 0}

from .simulation import Simulation
from .simulator import Simulator
from .simpv import SimPV
from .simps import SimPSTypeModel, SimPUTypeModel
from .simfactory import SimFactory

del simulation, simulator, simpv, simps
del simfactory
